/*1:*/
#line 16 "triecode.cw"

/*2:*/
#line 20 "triecode.cw"

#include "trie.h"

#include <stdio.h> 
#include <stdlib.h> 
#include <string.h> 
#include <assert.h> 

#define New(t)  (t*)malloc(sizeof(t))
#define NewV(t,s)  (t*)malloc(sizeof(t)*(s))

/*:2*/
#line 17 "triecode.cw"

/*3:*/
#line 33 "triecode.cw"

void abort(const char*msg){
fprintf(stderr,msg);
exit(1);
}

/*:3*//*4:*/
#line 43 "triecode.cw"

char check_char(char c){
if('a'<=c&&c<='z')
c-= 32;
if('A'<=c&&c<='Z')
return c;
return 0;
}

/*:4*//*5:*/
#line 54 "triecode.cw"

static TrieNode*trie_new_node(char c){
TrieNode*t= New(TrieNode);
if(!t)return 0;
t->data= c;
t->children= NULL;
t->n_child= 0;
t->size= 0;
t->is_terminal= 0;
return t;
}

/*:5*//*6:*/
#line 67 "triecode.cw"

Trie*trie_new(){
Trie*t= New(Trie);
if(!t)return 0;
t->n_words= 0;

t->root= trie_new_node(0);
if(!t->root)goto e1;

return t;
e1:
free(t);
return 0;
}

/*:6*//*7:*/
#line 84 "triecode.cw"

static TrieNode**new_child_array(int size){
TrieNode**nodes= NewV(TrieNode*,size);
if(!nodes)return 0;
int i;
for(i= 0;i<size;++i)
nodes[i]= NULL;
return nodes;
}

/*:7*//*8:*/
#line 96 "triecode.cw"

static TrieNode*in_children8(TrieNode*t,char c){
int i,n= t->n_child;
for(i= 0;i<n;++i)
if(t->children[i]->data==c)
return t->children[i];
return 0;
}

/*:8*//*9:*/
#line 106 "triecode.cw"

static TrieNode*add_children8(TrieNode*t,char c){
assert(t->n_child<8);
return t->children[t->n_child++]= trie_new_node(c);
}

/*:9*//*10:*/
#line 113 "triecode.cw"

static TrieNode*search_add_children26(TrieNode*t,char c,int insert){
assert(t->size==26);
int pos= c-'A';
if(t->children[pos])
return t->children[pos];
if(!insert)return 0;
TrieNode*node= t->children[pos]= trie_new_node(c);
if(!node)
abort("cannot add node in children26");
t->n_child++;
return node;
}

/*:10*//*11:*/
#line 128 "triecode.cw"

static TrieNode*in_children26(TrieNode*t,char c){
return search_add_children26(t,c,0);
}

static TrieNode*add_to_children26(TrieNode*t,char c){
return search_add_children26(t,c,1);
}

/*:11*//*12:*/
#line 139 "triecode.cw"

void add_existing_nodes26(TrieNode*t,TrieNode**nodes,int n){
assert(t->size==26);
int i;
for(i= 0;i<n;++i){
int pos= nodes[i]->data-'A';
assert(!t->children[pos]);
t->children[pos]= nodes[i];
}
}

/*:12*//*13:*/
#line 154 "triecode.cw"

static TrieNode*
get_or_insert_child_for_char(TrieNode*t,char c,int insert,int*modified,
int is_terminal){
assert('A'<=c&&c<='Z');
TrieNode*node;
/*14:*/
#line 169 "triecode.cw"

if(t->size==0){
if(!insert)return 0;

t->children= new_child_array(8);
if(!t->children)
abort("cannot add child array");
t->size= 8;
node= add_children8(t,c);
}
else if(t->size==8){

node= in_children8(t,c);
if(node)return node;

if(!insert)return 0;
if(t->n_child<8)
node= add_children8(t,c);
else{
TrieNode**current= t->children;
t->children= new_child_array(26);
if(!t->children)
abort("cannot add child array");
t->size= 26;
add_existing_nodes26(t,current,8);
node= add_to_children26(t,c);
}
}
else if(t->size==26){
node= in_children26(t,c);
if(node)return node;
if(!insert)return 0;
node= add_to_children26(t,c);
}else{
abort("wrong size of child list");
}

if(!node)
abort("cannot add node");
if(is_terminal)node->is_terminal= 1;
*modified= 1;
return node;
}

/*:14*/
#line 160 "triecode.cw"


/*:13*//*15:*/
#line 214 "triecode.cw"

static int
trie_visit_rec(TrieNode*t,char*word,int len,int insert,int*modified){
if(len==0){
if(*modified)return 2;
return 1;
}

char c= check_char(word[0]);
TrieNode*child= get_or_insert_child_for_char(
t,c,insert,modified,len==1);
if(!child){
if(insert)
abort("cannot insert word");
else return 0;
}
return trie_visit_rec(child,word+1,len-1,insert,modified);
}

/*:15*//*16:*/
#line 236 "triecode.cw"

int trie_visit(Trie*t,char*word,int insert){
int modified= 0;
int res= trie_visit_rec(t->root,word,strlen(word),insert,&modified);
if(insert&&res==2)t->n_words++;
return res;
}

/*:16*//*17:*/
#line 248 "triecode.cw"

static int trie_print_rec(TrieNode*t,char*word,int pos,int max_size){
if(pos==max_size){
word[pos]= '\0';
printf("word truncated: %s\n",word);
return 0;
}

if(t->is_terminal){
word[pos]= '\0';
printf("%s\n",word);
}

int i;
for(i= 0;i<t->size;++i){
TrieNode*node= t->children[i];
if(node){
word[pos]= node->data;
trie_print_rec(node,word,pos+1,max_size);
}
}
return 1;
}

#define MAX_SIZE 64

/*:17*//*18:*/
#line 275 "triecode.cw"

int trie_print(Trie*t){
printf("printing words\n");
char word[MAX_SIZE+1];
trie_print_rec(t->root,word,0,MAX_SIZE);
return 1;
}

/*:18*/
#line 18 "triecode.cw"

/*:1*/
